import asyncio
import inspect
import os
import urllib
from asyncio import Task
from typing import Optional

import pydantic

settings_module = None
modsloaded = False

def get_config(confvar, default=None):
  global settings_module, modsloaded
  o = os.environ.get(confvar)
  if o:
    return o
  if not modsloaded:
    try:
      settings_module = __import__('settings')
    except ModuleNotFoundError:
      settings_module = __import__('config')
    modsloaded = True
  if settings_module:
    if hasattr(settings_module,"get"):
      return settings_module.get(confvar, default)
    return getattr(settings_module, confvar, default)
  return default


connectors = None
odim_module = None

def get_base_from_module(module, parent_class):
  for n,x in inspect.getmembers(module, inspect.isclass):
    if issubclass(x, parent_class) and x!=parent_class and module.__name__ == x.__module__:
      return x


def get_connector_for_model(model):
  global connectors, odim_module
  if not connectors:
    connectors = [
      __import__('odim.mongo', fromlist=['odim']),
      __import__('odim.mysql', fromlist=['odim'])
    ]
    odim_module = __import__('odim')
  for connector in connectors:
    basemod = get_base_from_module(connector, pydantic.main.BaseModel)
    cls = model if inspect.isclass(model) else model.__class__
    if issubclass(cls, basemod):
      return get_base_from_module(connector, odim_module.Odim)

  if hasattr(model,'Config'):
    if hasattr(model.Config, 'db_name'):
      conn = get_connection_info(model.Config.db_name)
      if not conn:
        conn = get_connection_info(model.Config.db_uri)
      if conn:
        for connector in connectors:
          odim_class = get_base_from_module(connector, odim_module.Odim)
          if conn.protocol in odim_class.protocols:
            return odim_class

  raise AttributeError("No connector was found for instance class. Do you have the db_name or db_uri Config attr set?")



class ConnParams(pydantic.BaseModel):
  protocol : str
  host : str
  port : Optional[int] = None
  username : Optional[str] = None
  password : Optional[str] = None
  db : Optional[str]

  def url(self, withdb=True):
    u = self.protocol+"://"
    if self.username:
      u+= self.username+":"+self.password
    u+= self.host
    if self.port:
      u+= ":"+str(self.port)
    if withdb and self.db:
      u+= "/"+self.db
    return u


def get_connection_info(db) -> ConnParams:
  dbs = get_config('DATABASES')
  if db in dbs:
    if not isinstance(dbs[db], str):
      return ConnParams(**dbs[db])
    else:
      db = dbs[db]
  parsed = urllib.parse.urlparse(db)
  cp = ConnParams(protocol=parsed.scheme, host=parsed.hostname)
  if parsed.port:
    cp.port = parsed.port
  if parsed.username:
    cp.username = urllib.parse.unquote(parsed.username)
    cp.password = urllib.parse.unquote(parsed.password)
  if parsed.path:
    cp.db = parsed.path[1:]
  return cp



import nest_asyncio
nest_asyncio.apply()


def get_asyncio_loop(which=False):
  ''' Tryies to get either the running loop, wrapped loop or create a loop '''
  try:
    loop = asyncio.get_running_loop()
    existed = True
  except RuntimeError as e:  # no event loop running:
    if str(e) == 'no running event loop':
      loop = asyncio.new_event_loop()
      nest_asyncio.apply(loop)
      existed = False
    else:
      raise
  if which:
    return loop, existed
  return loop


def asyncio_run(future, as_task=True):
  """
  A better implementation of `asyncio.run`.

  :param future: A future or task or call of an async method.
  :param as_task: Forces the future to be scheduled as task (needed for e.g. aiohttp).
  """
  loop, existed = get_asyncio_loop(True)
  if existed:
    return asyncio.run(_to_task(future, as_task, loop))
  else:
    return loop.run_until_complete(_to_task(future, as_task, loop))


def _to_task(future, as_task, loop):
  if not as_task or isinstance(future, Task):
    return future
  return loop.create_task(future)


def awaited(o):
  while inspect.iscoroutine(o):
    loop = get_asyncio_loop()
    try:
      o = asyncio.run(o)
    except RuntimeError as e:
      # if "cannot be called from a running event loop" in str(e):
        o = loop.run_until_complete(o)
      # else:
      #   raise
    # o = asyncio_run(o, False)
  else:
    return o


import logging
import re
from datetime import datetime
from decimal import Decimal
from typing import List, Optional, Union

import bson
from bson import ObjectId as BsonObjectId, decimal128
from motor.motor_asyncio import AsyncIOMotorClient
from pydantic import BaseModel, Field
import inspect

from pymongo import ASCENDING, DESCENDING

from odim import BaseOdimModel, NotFoundException, Odim, Operation, SearchParams, all_json_encoders
from odim.helper import get_asyncio_loop, get_connection_info

log = logging.getLogger("uvicorn")

client_connections = {}


def get_mongo_client(alias):
  global client_connections
  if alias not in client_connections:
    cinf = get_connection_info(alias)
    conn = AsyncIOMotorClient(cinf.url(withdb=False), io_loop=get_asyncio_loop())
    client_connections[alias] = conn[cinf.db]
  return client_connections[alias]


class ObjectId(BsonObjectId):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not BsonObjectId.is_valid(v):
            raise ValueError('Invalid objectid')
        return BsonObjectId(v)

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(type='string')



class BaseMongoModel(BaseOdimModel):
  id: Optional[ObjectId] = Field(alias='_id', description="Unique identifier of the object") #

  class Config:
    arbitrary_types_allowed = True
    allow_population_by_field_name = True
    # underscore_attrs_are_private = True
    json_encoders = {
      ObjectId: str,
      BsonObjectId: str,
      datetime: lambda dt: dt.isoformat(),
      Decimal : float #im afraid im loosing precission here or we might output it as string?
    }

all_json_encoders.update( BaseMongoModel.Config.json_encoders)


def convert_decimal(dict_item):
  if dict_item is None:
    return None
  elif isinstance(dict_item, list):
    l = []
    for x in dict_item:
      l.append(convert_decimal(x))
    return l
  elif isinstance(dict_item, dict):
    nd = {}
    for k, v in list(dict_item.items()):
      nd[k] = convert_decimal(v)
    return nd
  elif isinstance(dict_item, Decimal):
    return bson.Decimal128(str(dict_item))
  else:
    return dict_item


class OdimMongo(Odim):
  protocols = ["mongo","mongodb"]

  def get_collection_name(self):
    if hasattr(self.model, 'Config'):
      if hasattr(self.model.Config, 'collection_name'):
        cn = self.model.Config.collection_name
        return cn
    return self.model.__class__.__name__


  async def get_mongo_client(self):
    return get_mongo_client(self.get_connection_identifier())


  def mongo(self):
    return self.get_mongo_client()[self.get_collection_name()]


  async def get(self, id : Union[str, ObjectId], extend_query : dict= {}, include_deleted : bool = False):
    if isinstance(id, str):
      id = ObjectId(id)
    mongo_client = await self.get_mongo_client()
    collection = self.get_collection_name()
    softdel = {self.softdelete(): False} if self.softdelete() and not include_deleted else {}
    ret = await mongo_client[collection].find_one({"_id" : id, **softdel, **self.get_parsed_query(extend_query)})
    if not ret:
      raise NotFoundException()
    ret = self.execute_hooks("pre_init", ret)
    x = self.model(**ret)
    return self.execute_hooks("post_init", x)


  async def save(self, extend_query : dict= {}, include_deleted : bool = False) -> ObjectId:
    mongo_client = await self.get_mongo_client()
    if not self.instance:
      raise AttributeError("Can not save, instance not specified ")#describe more how ti instantiate
    collection = self.get_collection_name()
    iii = self.execute_hooks("pre_save", self.instance, created=(not self.instance.id))
    dd = convert_decimal(iii.dict(by_alias=True))

    if not self.instance.id:
      dd["_id"] = BsonObjectId()
      iii = dd["_id"]
      self.instance.id = dd["_id"]
      softdel = {self.softdelete(): False} if self.softdelete() else {}
      ret = await mongo_client[collection].insert_one({**dd, **extend_query, **softdel})
      created = True
    else:
      softdel = {self.softdelete(): False} if self.softdelete() and not include_deleted else {}
      ret = await mongo_client[collection].replace_one({"_id" : self.instance.id, **softdel, **self.get_parsed_query(extend_query)}, dd)
      assert ret.modified_count > 0, "Not modified error"
      created = False
    iii = self.execute_hooks("post_save", iii, created=created)
    return self.instance.id


  async def update(self, extend_query : dict= {}, include_deleted : bool = False, only_fields : Optional[List['str']] = None):
    ''' Saves only the changed fields leaving other fields alone '''
    mongo_client = await self.get_mongo_client()
    collection = self.get_collection_name()
    iii = self.execute_hooks("pre_save", self.instance)
    dd = convert_decimal(iii.dict(exclude_unset=True, by_alias=True))
    if "_id" not in dd:
      raise AttributeError("Can not update document without _id")
    dd_id = dd["_id"]
    if isinstance(dd_id, str):
      dd_id = ObjectId(dd_id)
    del dd["_id"]
    if only_fields and len(only_fields)>0:
      dd = dict([(key, val) for key, val in dd.items() if key in only_fields])
    softdel = {self.softdelete(): False} if self.softdelete() and not include_deleted else {}
    ret = await mongo_client[collection].find_one_and_update({"_id" : dd_id, **softdel, **self.get_parsed_query(extend_query)}, {"$set" : dd})
    iii = self.execute_hooks("post_save", iii, created=False)
    return ret


  def get_parsed_query(self, query):
    rsp = {}
    for k, (op, v) in self.parse_query_operations(query).items():
      #perhaps use model to ensure the search value is of correct type
      if op == Operation.exact:
        rsp[k] = v
      elif op == Operation.isnot:
        rsp[k] = { "$ne" : v}
      elif op == Operation.contains:
        rsp[k] = { "$regex" : '.*'+str(v)+'.*', "$options" : "i" }
      elif op == Operation.gt:
        rsp[k] = { "$gt" : v}
      elif op == Operation.gte:
        rsp[k] = { "$gte" : v}
      elif op == Operation.lt:
        rsp[k] = { "$lt" : v}
      elif op == Operation.gt:
        rsp[k] = { "$lte" : v}
      elif op == Operation.null:
        if v:
          rsp["$or"] = [ {k : {"$exists" : False}}, {k : None} ]
        else:
          rsp["$and"] = [ {k : {"$exists" : True}}, {k: { "$ne" : None }} ]
    return rsp


  async def find(self, query : dict, params : SearchParams = None, include_deleted : bool = False):
    mongo_client = await self.get_mongo_client()
    collection = self.get_collection_name()
    if self.softdelete() and not include_deleted:
      query = {self.softdelete(): False, **query}
    #TODO use projection on model to limit to only desired fields
    find_params = {}
    if params:
      find_params["skip"] = params.offset
      find_params["limit"] = params.limit
      if params.sort not in (None,''):
        find_params["sort"] = []
        for so in params.sort.split(','):
          if so[0] == "-":
            find_params["sort"].append( (so[1:], DESCENDING) )
          else:
            find_params["sort"].append( (so, ASCENDING) )
    query = self.get_parsed_query(query)
    curs = mongo_client[collection].find(query, **find_params)
    rsplist = []
    for x in await curs.to_list(None):
      x2 = self.execute_hooks("pre_init", x)
      m = self.model( **x2 )
      rsplist.append( self.execute_hooks("post_init", m) )
    return rsplist



  async def count(self, query : dict, include_deleted : bool = False):
    if self.softdelete() and not include_deleted:
      query = {self.softdelete(): False, **query}
    mongo_client = await self.get_mongo_client()
    collection = self.get_collection_name()
    return await mongo_client[collection].count_documents(query)


  async def delete(self, obj : Union[str, ObjectId, BaseMongoModel], extend_query : dict= {}, force_harddelete : bool = False):
    mongo_client = await self.get_mongo_client()
    collection = self.get_collection_name()
    if isinstance(obj, str):
      d = {"_id" : ObjectId(obj)}
    elif isinstance(obj, ObjectId):
      d = {"_id" : obj}
    else:
      d = obj.dict()
    d.update(self.get_parsed_query(extend_query))
    softdelete = self.softdelete() and not force_harddelete
    if self.has_hooks("pre_remove","post_remove"):
      ret = await mongo_client[collection].find_one(d)
      if not ret:
        raise NotFoundException()
      ret = self.execute_hooks("pre_init", ret)
      x = self.model(**ret)
      x = self.execute_hooks("post_init", x)
      x = self.execute_hooks("pre_remove", x, softdelete=softdelete)
    if softdelete:
      rsp = await mongo_client[collection].find_one_and_update(d, {"$set" : {self.softdelete() : True}})
    else:
      rsp = await mongo_client[collection].delete_one(d)
    if self.has_hooks("post_remove"):
      self.execute_hooks("post_remove", x, softdelete=softdelete)
    return rsp


'''
Contains the extended FastAPI router, for simplified CRUD from a model
'''
from typing import Any, List, Optional, Sequence, Set, Type, Union

import fastapi
from fastapi import Depends, params
from pydantic import BaseModel, create_model

from odim import Odim, OkResponse, SearchResponse
from odim.dependencies import SearchParams

class OdimRouter(fastapi.APIRouter):
  ''' Simplified FastAPI router for easy CRUD '''

  def mount_crud(self,
                 path: str,
                 *,
                 model : Type[BaseModel],
                 tags: Optional[List[str]] = None,
                 dependencies : Optional[Sequence[params.Depends]] = None,
                 include_in_schema: bool = True,
                 methods : Optional[Union[Set[str], List[str]]] = ('create','get','search','save','update','delete'),
                 methods_exclude : Optional[Union[Set[str], List[str]]] = [],
                 extend_query : dict= {}):
    ''' Add endpoints for CRUD operations for particular model
    :param path: base_path, for the model resource location eg: /api/houses/
    :param model: pydantic/Odim BaseModel, that is used for eg. Houses
    :param tags: Starlette/FastAPI tags for endpoints
    :param dependencies: Starlette/FastAPI dependencies for all endpoints
    :param include_in_schema: whether to include in docs
    :param methods: methods to automatically generate ('create','get','search','save','update','delete')
    :param methods_exclude: methods to NOT automatically generate ('create','get','search','save','update','delete')
    :param extend_query: adds these parameters to every query and sets it on the object upon creation. keys are fields, values can be exact values or functions taking request as parameter
    '''
    add_methods = [ x for x in methods if x not in methods_exclude ]

    if 'create' in add_methods:
      async def create(request : fastapi.Request, obj : model):
        for k, v in exec_extend_qeury(request,extend_query).items():
          setattr(obj, k, v)
        await Odim(obj).save()
        return obj
      self.add_api_route(path = path,
                         endpoint=create,
                         response_model=model,
                         status_code=201,
                         tags=tags,
                         dependencies = dependencies,
                         summary="Create new %s" % model.schema().get('title'),
                         description = "Create new instance of %s " %  model.schema().get('title'),
                         methods = ["POST"],
                         include_in_schema = include_in_schema)

    if 'get' in add_methods:
      async def get(request : fastapi.Request, id : str):
        return await Odim(model).get(id=id, extend_query=exec_extend_qeury(request,extend_query))
      self.add_api_route(path = path+"{id}",
                         endpoint=get,
                         response_model=model,
                         tags=tags,
                         dependencies = dependencies,
                         summary="Get %s by id" % model.schema().get('title'),
                         description = "Return individual %s details " % model.schema().get('title'),
                         methods = ["GET"],
                         include_in_schema = include_in_schema)

    if 'search' in add_methods:
      async def search(request : fastapi.Request, search_params : dict = Depends(SearchParams)):
        sp = {**search_params.q, **exec_extend_qeury(request,extend_query)}
        rsp = { "results" : await Odim(model).find(sp, search_params),
                "total" : await Odim(model).count(sp),
                "search" : search_params.dict()}
        return rsp
      self.add_api_route(path = path,
                         endpoint=search,
                         response_model=SearchResponse[model],
                         tags=tags,
                         dependencies = dependencies,
                         summary="Search for %ss" % model.schema().get('title'),
                         description = "Performs a listing search for %s " %  model.schema().get('title'),
                         methods = ["GET"],
                         include_in_schema = include_in_schema)

    if 'save' in add_methods:
      async def save(request : fastapi.Request, id : str, obj : model):
        obj.id = id
        await Odim(obj).save(extend_query=exec_extend_qeury(request,extend_query))
        return obj
      self.add_api_route(path = path+"{id}",
                     endpoint=save,
                     response_model=model,
                     tags=tags,
                     dependencies = dependencies,
                     summary="Replace %s by id" % model.schema().get('title'),
                     description = "PUT replaces the original %s as whole  " %  model.schema().get('title'),
                     methods = ["PUT"],
                     include_in_schema = include_in_schema)

    if 'update' in add_methods:
      async def update(request : fastapi.Request, id : str, obj : model):
        obj.id = id
        await Odim(obj).update(extend_query=exec_extend_qeury(request,extend_query))
        return obj
      self.add_api_route(path = path+"{id}",
                     endpoint=update,
                     response_model=model,
                     tags=tags,
                     dependencies = dependencies,
                     summary="Partial update %s by id" % model.schema().get('title'),
                     description = "Just updates individual fields of %s " %  model.schema().get('title'),
                     methods = ["Patch"],
                     include_in_schema = include_in_schema)

    if 'delete' in add_methods:
      async def delete(request : fastapi.Request, id : str) -> None:
        await Odim(model).delete(id, extend_query=exec_extend_qeury(request,extend_query))
        return OkResponse()
      self.add_api_route(path = path+"{id}",
                     endpoint=delete,
                     response_model=OkResponse,
                     status_code=200,
                     tags=tags,
                     dependencies = dependencies,
                     summary="Delete %s by id" % model.schema().get('title'),
                     description = "Deletes individual instance of %s " %  model.schema().get('title'),
                     methods = ["DELETE"],
                     include_in_schema = include_in_schema)



  def generate(self,
                 path: str,
                 *,
                 model : Type[BaseModel],
                 tags: Optional[List[str]] = None,
                 dependencies : Optional[Sequence[params.Depends]] = None,
                 include_in_schema: bool = True,
                 methods : Optional[Union[Set[str], List[str]]] = ('create','get','search','save','update','delete'),
                 methods_exclude : Optional[Union[Set[str], List[str]]] = []):
    ''' Generates the code for the endpoints
    :param path: base_path, for the model resource location eg: /api/houses/
    :param model: pydantic/Odim BaseModel, that is used for eg. Houses
    :param tags: Starlette/FastAPI tags for endpoints
    :param dependencies: Starlette/FastAPI dependencies for all endpoints
    :param include_in_schema: whether to include in docs
    :param methods: methods to automatically generate ('create','get','search','save','update','delete')
    :param methods_exclude: methods to NOT automatically generate ('create','get','search','save','update','delete')
    '''
    add_methods = [ x for x in methods if x not in methods_exclude ]
    model_name = model.__name__
    other=""
    if tags:
      other+= ", tags="+str(tags)
    if dependencies:
      other+= ", dependencies="+str(dependencies)
    if not include_in_schema:
      other+= ", include_in_schema=False"

    if 'get' in add_methods:
      print(f'''
@router.get("{path}{{id}}", response_model={model_name}{other})
async def get_{model_name}(id : str):
  \'\'\' Returns the individual {model_name} details\'\'\'
  return await Odim(TestTable).get(id=id)
''')

    if 'search' in add_methods:
      print(f'''
@router.get("{path}", response_model=SearchResponse[{model_name}]{other})
async def search_{model_name}(search : dict = Depends(SearchParams)):
  rsp = {{ "results" : await Odim({model_name}).find(search.q, search),
          "total" : await Odim({model_name}).count(search.q),
          "search" : search.dict()}}
  return rsp
''')


    if 'create' in add_methods:
      print(f'''
@router.post("{path}", status_code=201, response_model={model_name}{other})
async def create_{model_name}(obj : {model_name}):
  await Odim(obj).save()
  return obj
''')

    if 'save' in add_methods:
      print(f'''
@router.put("{path}{{id}}", response_model={model_name}{other})
async def save_{model_name}(id : str, obj : {model_name}):
  obj.id = id
  await Odim(obj).save()
  return obj
''')

    if 'update' in add_methods:
      print(f'''
@router.patch("{path}{{id}}", response_model={model_name}{other})
async def update_{model_name}(id : str, obj : {model_name}):
  obj.id = id
  await Odim(obj).update()
  return obj
''')

    if 'delete' in add_methods:
      print(f'''
@router.delete("{path}{{id}}", status_code=200, response_model=OkResponse)
async def delete_{model_name}(id : str):
  await Odim(obj).delete(id)
  return OkResponse()
''')



def exec_extend_qeury(request : fastapi.Request, sl : dict = {}):
  out = {}
  for k, v in sl.items():
    if callable(v):
      out[k] = v(request)
    else:
      out[k] = v
  return out


pre_init = 'pre_init'
post_init = 'post_init'
pre_validate = 'pre_validate'
post_validate = 'post_validate'
pre_save = 'pre_save'
post_save = 'post_save'
pre_remove = 'pre_remove'
post_remove = 'post_remove'


import inspect
import json
import random
import string
import importlib
import sys
from decimal import Decimal
from enum import Enum
from os import path, getcwd
from typing import Any, List, Optional, Type, Union
from odim.basesignals import BaseSignals
from pydantic import BaseModel, Field, create_model
from odim.mongo import BaseMongoModel, ObjectId
from datetime import datetime
from odim import dynmodels

def get_class_by_name(classname):
  ''' Returns the loaded instance of a class '''
  mod_name = classname[:classname.rindex('.')]
  module = __import__(mod_name, fromlist=[mod_name, ])
  the_class = getattr(module, classname[classname.rindex('.') + 1:])
  return the_class


MM_TYPE_MAPPING = {
  "String" : 'str',
  "Number" : 'float',
  "Boolean" : "bool",
  "Array" : "List",
  "List" : "List",
  "Date" : "datetime",
  "Int" : "int",
  "Decimal" : "Decimal",
  "Decimal128" : "Decimal",
  "ObjectId" : "ObjectId",
  "Enum" : "str",
  "Mixed" : "Any"}


DM_TYPE_MAPPING = {
  "String" : str,
  "Number" : int,
  "Boolean" : bool,
  "Array" : list,
  "List" : list,
  "Date" : datetime,
  "ObjectId" : ObjectId,
  "Parent": dict,
  "Int" : int,
  "Decimal" : Decimal,
  "Decimal128" : Decimal,
  "Enum" : str,
  "Mixed" : Any
}

class SEnum(str, Enum):
  pass

def encode(k, v):
  if isinstance(v, list):
    if len(v) == 0:
      return Optional[List[Any]], None
    else:
      if v[0] == "Mixed": #workaround for subdocuments
        return Optional[List[Any]], None
      else:
        enc = encode("sub", v[0])
        return Optional[List[enc[0]]], enc[1]

  elif isinstance(v, dict):
      if v.get("type") == "Parent":
        subcls = {}
        for ks,vs in v.get("child", {}).items():
          subcls[ks] = encode(ks, vs)
        m = create_model(__model_name=v.get("__title", k+(''.join(random.choices(string.ascii_uppercase + string.digits, k=6)))).capitalize(), #TODO have a register for duplicate class names or use already made class
                         __module__ = "odim.dynmodels",
                         __base__=BaseModel,
                         **subcls)
        if "__description" in v:
          m.__doc__ = v.get("__description")
        dt = m
      elif v.get("type") == "Enum":
        subcls = {}
        for opt in v.get("options",[]):
          subcls[opt] = opt

        m = SEnum(v.get("__title", k.capitalize()+"Enum"+(''.join(random.choices(string.ascii_uppercase + string.digits, k=6)))), subcls)
        if "__description" in v:
          m.__doc__ = v.get("__description")
        dt = m
      else:
        dt = DM_TYPE_MAPPING.get(v.get("type"), str)

      field = Field(description=v.get("__description",v.get("description","")), title=v.get("__title", v.get("title")))
      for sk, sv in v.items():
        if sk not in ("type","child","parent","description","__description","title","__title","required","default_factory","const","alias"):
          if sk in ('gt','ge','lt','le','multiple_of','min_items','max_items','min_length','max_length','regex'):
            setattr(field, sk, sv)
          else:
            field.extra[sk] =  sv
      if v.get("required", False) or v.get("default", False) not in ('', False, None):
        field.default = v.get("default",...) #TODO default value removes the required attribute
        return dt, field
      else:
        field.default = v.get("default", None)
        return Optional[dt], field
  else: #a string is there as value
    return Optional[DM_TYPE_MAPPING.get(v, str)], None


def location_tester(file_uri):
  tryfiles = [file_uri,
              path.join(getcwd(), file_uri),
              path.join(getcwd(), "models", file_uri),
              path.join(path.dirname(path.realpath(__file__)), file_uri)]
  for f in tryfiles:
    if path.exists(f):
      return f



class ModelFactory(object):
  ''' Utility  for  generating stub code for Pydantic models based on their JSON definition and vice-versa'''

  @classmethod
  def load_mongo_model(cls, class_name=None,
                       description=None,
                       db_name=None, db_uri=None,
                       database=None, collection_name=None,
                       softdelete=None,
                       file_uri=None, signal_file=None) -> Type[BaseMongoModel]:

    assert db_name or db_uri, "Either database_name or database_uri must be specified"
    assert database and collection_name, "database and collection_name must be set"
    if not file_uri:
      file_uri = "schemas/src/"+database+"/"+collection_name.lower() +".json"
    file = location_tester(file_uri)
    assert file, "No schema json was found."

    if not signal_file:
      signal_file = "schemas/dist/python3/odim/hooks/"+database+"/"+collection_name.lower() +".py"
    signal_file = location_tester(signal_file)

    with open(file, "r") as f:
      data = json.loads(f.read())
      newcls = {}
      for k,v in data.items():
        if k in ("__class_name","__title"):
          if not class_name:
            class_name = v
        elif k in ("__description"):
          if not description:
            description = v
        else:
          newcls[k] = encode(k, v)

      if not class_name:
        class_name = collection_name
      m = create_model(class_name,
                       __module__ = "odim.dynmodels",
                       __base__=BaseMongoModel,
                       **newcls)
      meta_attrs = {"collection_name": collection_name, **vars(BaseMongoModel.Config)}
      if db_name:
        meta_attrs["db_name"] = db_name
      if db_uri:
        meta_attrs["db_uri"] = db_uri
      if softdelete:
        meta_attrs["softdelete"] = softdelete
      if signal_file: # now handle the signals
        spec = importlib.util.spec_from_file_location("odim.dynmodels.%s.signals" % class_name, signal_file)
        foo = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(foo)
        for n,x in inspect.getmembers(foo):
          if inspect.isfunction(x):
            if not "odim_hooks" in meta_attrs:
              meta_attrs["odim_hooks"] = {"pre_init":[], "post_init":[], "pre_save":[], "post_save":[],"pre_remove":[],"post_remove":[],"pre_validate":[],"post_validate":[]}
            if n in meta_attrs["odim_hooks"].keys():
              meta_attrs["odim_hooks"][n].append(x)
          elif inspect.isclass(x) and issubclass(x, BaseSignals) and x!=BaseSignals:
            for cfn,cfx in inspect.getmembers(x, predicate=inspect.ismethod):
              if not getattr(cfx,'__isabstractmethod__',None):
                if not "odim_hooks" in meta_attrs:
                  meta_attrs["odim_hooks"] = {"pre_init":[], "post_init":[], "pre_save":[], "post_save":[],"pre_remove":[],"post_remove":[],"pre_validate":[],"post_validate":[]}
                if cfn in meta_attrs["odim_hooks"].keys():
                  meta_attrs["odim_hooks"][n].append(cfx)

      setattr(m, 'Config', type('class', (), meta_attrs))
      m.__doc__ = description
      m.update_forward_refs()
      return m


  @classmethod
  def model_to_json(cls, model : Union[str, BaseModel, BaseModel.__class__]):
    if isinstance(model, str):
      newcls = get_class_by_name(model)
    elif not inspect.isclass(model):
      newcls = model.__class__
    else:
      newcls = model
    pydschema = newcls.schema()
    out = {}
    for propname, vals in pydschema["properties"].items():
      out[propname] = {}
      if vals["type"] == "string" and vals.get("format") == "date-time":
        out[propname]["type"] = "Date"
      if vals["type"] == "string" and propname.endswith("_id"):
        out[propname]["type"] = "ObjectId"
      elif vals["type"] == "string":
        out[propname]["type"] = "String"
      elif vals["type"] == "boolean":
        out[propname]["type"] = "Boolean"
      elif vals["type"] in ("integer","number"):
        out[propname]["type"] = "Number"
      elif vals["type"].lower() in ("float","double",'decimal'):
        out[propname]["type"] = "Decimal128"
      elif vals["type"] == "array":
        out[propname]["type"] = "Array"

      if "default" in vals:
        out[propname]["default"] = vals["default"]
      if "description" in vals:
        out[propname]["description"] = vals["description"]
    print(json.dumps(out, indent=4))


  @classmethod
  def json_to_fields(cls, js_data):
    if js_data.endswith(".js") or js_data.endswith(".json"):
      filename = js_data[js_data.rindex("/")+1:] if "/" in js_data else js_data
      filename = filename.replace(".json","").replace(".js","")
      with open(js_data, "r") as f:
        data = json.loads(f.read())
    else:
      filename = None
      data = json.loads(js_data)

    if filename:
      print(f"class {filename}(BaseMongoModel):")
    for k,v in data.items():
      if not isinstance(v, dict):
        print(f"  {k} : Optional[{MM_TYPE_MAPPING[v]}]")
      else:
        if v.get("type") == "Number" and v.get("integer"):
          dt = "int"
        else:
          dt = MM_TYPE_MAPPING.get(v.get("type"), "str")
        if v.get("required", False) or v.get("default", False):
          dt = f"Optional[{dt}]"
        rest = ""
        if "description" in v:
          rest+= f" = Field(description='{v['description']}')"
        print(f"  {k} : {dt} {rest}")




if __name__ == "__main__":
  import os.path, sys
  sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
  if len(sys.argv) != 2:
    print("The script requires 1 parameter and that is either the package full name, or json file path")
    sys.exit()
  if sys.argv[1].endswith(".js") or sys.argv[1].endswith(".json"):
    ModelFactory.json_to_fields(sys.argv[1])
  else:
    ModelFactory.model_to_json(sys.argv[1])


import json
from typing import Optional
from fastapi import Query
from pydantic import BaseModel, Field, ValidationError
import re

class SearchParams:

  def __init__( self,
                q: Optional[str] = Query(None, description="Query string to search. Urlencoded key=value or JSON dictionary with fields"),
                limit: Optional[int] = Query(description="Limit the number of results to x", default=25),
                offset: Optional[int] = Query(description="From which record to start", default=0),
                sort: Optional[str] = Query(None, description="Fields to sort by. Comma separated list of fields (- determines DESC order). ", example="name,-created_at")
              ):
    self.limit = limit
    self.offset = offset
    self.sort = sort
    if q:
      try:
        self.q = json.loads(q)
      except:
        if "=" in q:
          key, value = q.split("=")
          self.q = {key : value}
        else:
          raise ValidationError("q parameter must be either a urlencoded JSON or key=value entry")
      for key in self.q:
        if not re.match("[a-zA-Z0-9_]+", key):
          raise ValidationError("The field in q param seems to be incorrect")
    else:
      self.q = {}


  def __getitem__(self, item):
    return getattr(self, item)


  def dict(self):
    return {"q":self.q, "limit":self.limit, "offset":self.offset, "sort":self.sort}

  def __str__(self):
    return json.dumps(self.dict())


from abc import abstractmethod


class BaseSignals:

  @classmethod
  @abstractmethod
  def pre_init(cls, sender, doc, *args, **kwargs):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def post_init(cls, sender, instance, *args, **kwargs):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def pre_save(cls, sender, instance, *args, **kwargs):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def post_save(cls, sender, instance, created, *args, **kwargs):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def pre_validate(cls, sender, instance, *args, **kwargs):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def post_validate(cls, sender, instance, *args, **kwargs):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def pre_remove(cls, sender, instance, softdelete, *args, **kwargs):
    raise NotImplementedError()

  @classmethod
  @abstractmethod
  def post_remove(cls, sender, instance, softdelete,  *args, **kwargs):
    raise NotImplementedError()